package quimufu.colourful_portals.portal;

import com.ibm.icu.impl.Pair;
import net.minecraft.block.BlockState;
import net.minecraft.block.Blocks;
import net.minecraft.entity.Entity;
import net.minecraft.registry.RegistryKey;
import net.minecraft.registry.RegistryKeys;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.DyeColor;
import net.minecraft.util.Identifier;
import net.minecraft.util.TypeFilter;
import net.minecraft.util.math.*;
import net.minecraft.world.World;
import org.jetbrains.annotations.Nullable;
import qouteall.imm_ptl.core.api.PortalAPI;
import qouteall.imm_ptl.core.portal.Portal;
import qouteall.q_misc_util.my_util.DQuaternion;
import quimufu.colourful_portals.config.ColourfulPortalConfig;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import static quimufu.colourful_portals.ColourfulPortalsMod.*;
import static quimufu.colourful_portals.Components.PORTAL_CANDIDATE_LIST;
import static quimufu.colourful_portals.Components.PORTAL_LIST;

public class PortalManager {

    public static void onPortalBlockPlaced(ServerWorld world, BlockPos pos, Identifier blockId) {
        world.getProfiler().push("onPortalBlockPlaced");
        LOGGER.debug("onPortalBlockPlaced, {}", blockId);

        PortalListComponent portalCandidateList = PORTAL_CANDIDATE_LIST.get(world.getLevelProperties());
        //delete portalCandidates obstructed by PortalBlocks for consistency
        List<BlockBox> portalCandidates = portalCandidateList.getContainingPortals(blockId, pos, world.getDimensionKey().getValue());
        LOGGER.debug("got containing PortalCandidates , {}", portalCandidates);
        for (BlockBox portalCandidate : portalCandidates) {
            if (!PortalHelper.isValidCandidate(world, portalCandidate, blockId)) {
                LOGGER.debug("invalid, {}", portalCandidate);
                Pair<BlockBox, Identifier> portalWithDim =
                        Pair.of(portalCandidate, world.getDimensionKey().getValue());

                portalCandidateList.removePortal(blockId, portalWithDim);
            }
        }
        //find new portalCandidates created by PortalBlock placement
        List<BlockBox> portals = PortalHelper.findPortalCandidates(world, pos, blockId);
        LOGGER.debug("new portalCandidates found, {}", portals);

        for (BlockBox portal : portals) {
            Pair<BlockBox, Identifier> portalWithDim = Pair.of(portal, world.getDimensionKey().getValue());
            portalCandidateList.createPortal(blockId, portalWithDim);
        }
        world.getProfiler().pop();
    }

    public static void onPortalBlockBroken(ServerWorld world, BlockPos pos, Identifier blockId) {
        world.getProfiler().push("onPortalBlockBroken");
        LOGGER.debug("onPortalBlockBroken, {}", blockId);

        PortalListComponent portalList = PORTAL_LIST.get(world.getLevelProperties());
        PortalListComponent portalCandidateList = PORTAL_CANDIDATE_LIST.get(world.getLevelProperties());

        //check portalCandidate validity
        List<BlockBox> portalCandidates = portalCandidateList.getContainingPortals(blockId, pos, world.getDimensionKey().getValue());
        LOGGER.debug("getContainingPortals, {}", portalCandidates);
        for (BlockBox portalCandidate : portalCandidates) {
            if (!PortalHelper.isValidCandidate(world, portalCandidate, blockId)) {
                LOGGER.debug("invalid, {}", portalCandidate);
                Pair<BlockBox, Identifier> portalWithDim =
                        Pair.of(portalCandidate, world.getDimensionKey().getValue());
                portalCandidateList.removePortal(blockId, portalWithDim);
            }
        }

        //add portalCandidates deobstructed by PortalBlock removal
        for (Direction direction : Direction.values()) {
            List<BlockBox> newPortalCandidates = PortalHelper.findPortalCandidates(world, pos.add(direction.getVector()), blockId);

            for (BlockBox newPortal : newPortalCandidates) {
                LOGGER.debug("potentially new candidate after deobstruction , {}", newPortal);

                PortalListComponent portalListComponent = PORTAL_CANDIDATE_LIST.get(world.getLevelProperties());
                Pair<BlockBox, Identifier> portalWithDim =
                        Pair.of(newPortal, world.getDimensionKey().getValue());
                portalListComponent.createPortal(blockId, portalWithDim);
            }
        }


        //check portal validity
        List<BlockBox> portals = portalList.getContainingPortals(blockId, pos, world.getDimensionKey().getValue());
        boolean changed = false;
        for (BlockBox portal : portals) {
            if (!PortalHelper.isValidPortal(world, portal, blockId)) {
                LOGGER.debug("portal became invalid ,{} {}", world.getDimensionKey().getValue(), portal);

                Pair<BlockBox, Identifier> portalWithDim =
                        Pair.of(portal, world.getDimensionKey().getValue());
                portalList.removePortal(blockId, portalWithDim);
                destroyPortalEntitiesInside(world, Box.from(portal));
                removePortalBlocks(world, portal);
                changed = true;
            }

        }
        if (changed) {
            fixPortalGroup(blockId, portalList, world.getServer());
        }

        world.getProfiler().pop();
    }

    private static void removePortalBlocks(ServerWorld world, BlockBox portal) {
        BlockPos.stream(portal)
                .filter(blockPos -> world.getBlockState(blockPos).isOf(PORTAL_BLOCK))
                .forEach(blockPos -> world.setBlockState(blockPos, Blocks.AIR.getDefaultState()));
    }

    private static void fixPortalGroup(Identifier blockId, PortalListComponent portalList, MinecraftServer server) {
        List<Pair<BlockBox, Identifier>> portalsWithDim = portalList.getPortals(blockId);
        if (portalsWithDim.size() == 1) {
            Pair<BlockBox, Identifier> portal = portalsWithDim.get(0);
            ServerWorld world = getPortalWorld(portal, server);
            if (world == null) {
                LOGGER.error("error fixing portalGroup, world {} was null", portal.second);
                return;
            }
            Pair<BlockBox, Identifier> portalWithDim = Pair.of(portal.first, portal.second);
            portalList.removePortal(blockId, portalWithDim);
            destroyPortalEntitiesInside(world, Box.from(portal.first));
            removePortalBlocks(world, portal.first);
            return;
        }
        for (int i = 0; i < portalsWithDim.size(); i++) {
            Pair<BlockBox, Identifier> portal = portalsWithDim.get(i);
            Pair<BlockBox, Identifier> linkedToPortal = i + 1 < portalsWithDim.size() ? portalsWithDim.get(i + 1) : portalsWithDim.get(0);
            assureLinkedTo(portal, linkedToPortal, server);
            assurePortalBlocksPlaced(blockId, portal, server);
        }
    }

    private static void assurePortalBlocksPlaced(Identifier blockId, Pair<BlockBox, Identifier> portal, MinecraftServer server) {
        DyeColor color = ColourfulPortalConfig.portalBlocks.get(blockId.toString());
        if (color != null) {
            Direction.Axis portalPlaneAxis = PortalHelper.getAxisW(portal.first)
                    .rotateClockwise(Direction.Axis.Y)
                    .getAxis();
            ServerWorld portalWorld = getPortalWorld(portal, server);
            if (portalWorld == null) {
                LOGGER.error("error placing portal planes, world {} was null", portal.second);
                return;
            }
            BlockState portalBlockState = PORTAL_BLOCK.getStateWith(color, portalPlaneAxis);

            PortalHelper.insideOf(portal.first)
                    .forEachRemaining(blockPos -> {
                        BlockState blockState = portalWorld.getBlockState(blockPos);
                        if (blockState.isAir() || blockState.getFluidState().isOf(PORTAL_FLUID)) {
                            portalWorld.setBlockState(blockPos, portalBlockState);
                        }
                    });
        }

    }

    private static void assureLinkedTo(Pair<BlockBox, Identifier> fromPortal, Pair<BlockBox, Identifier> linkedToPortal, MinecraftServer server) {

        ServerWorld fromPortalWorld = getPortalWorld(fromPortal, server);
        RegistryKey<World> linkedToPortalWorldRegKey = RegistryKey.of(RegistryKeys.WORLD, linkedToPortal.second);

        BlockBox fromPortalBlockBox = fromPortal.first;
        Box fromPortalBox = Box.from(fromPortalBlockBox);
        BlockBox linkedToPortalBlockBox = linkedToPortal.first;
        Box linkedToPortalBox = Box.from(linkedToPortalBlockBox);
        if (fromPortalWorld == null) {
            LOGGER.error("error linking portals, world {} was null", fromPortalBlockBox);
            return;
        }

        List<Portal> portals = fromPortalWorld.getEntitiesByType(TypeFilter.instanceOf(Portal.class), fromPortalBox, Entity::isAlive);
        List<Portal> outgoingPortals = new ArrayList<>();
        for (Portal portal : portals) {
            Box portalBoundingBox = portal.getBoundingBox();
            //is portal actually inside portalBox?
            if (fromPortalBox.contains(portalBoundingBox.minX, portalBoundingBox.minY, portalBoundingBox.minZ)
                    && fromPortalBox.contains(portalBoundingBox.maxX, portalBoundingBox.maxY, portalBoundingBox.maxZ)) {
                outgoingPortals.add(portal);
            }
        }
        if (outgoingPortals.size() > 2) {
            LOGGER.warn("Found more then 2 portals in {}, cleaning up", fromPortalBox);
            for (int i = 2; i < outgoingPortals.size(); i++) {
                Portal outgoingPortal = outgoingPortals.get(i);
                outgoingPortal.kill();
            }
        }
        if (outgoingPortals.isEmpty()) {
            Portal portal = Portal.entityType.create(fromPortalWorld);
            if (portal == null) {
                LOGGER.error("could not create Portal entity for  {}", fromPortal);
                return;
            }

            portal.setOriginPos(fromPortalBox.getCenter());
            portal.setDestinationDimension(linkedToPortalWorldRegKey);
            portal.setDestination(linkedToPortalBox.getCenter());
            Vec3d axisW = Vec3d.of(PortalHelper.getAxisW(fromPortalBlockBox).getVector());
            portal.setOrientationAndSize(
                    axisW, // axisW
                    new Vec3d(0, 1, 0), // axisH
                    2, // width
                    3 // height
            );
            portal.setRotationTransformation(DQuaternion.getRotationBetween(axisW, Vec3d.of(PortalHelper.getAxisW(linkedToPortalBlockBox).getVector())));
            outgoingPortals.add(portal);
            PortalAPI.spawnServerEntity(portal);
        }
        if (outgoingPortals.size() == 1) {
            Portal portal = PortalAPI.createFlippedPortal(outgoingPortals.get(0));
            outgoingPortals.add(portal);
            PortalAPI.spawnServerEntity(portal);
        }
        for (Portal outgoingPortal : outgoingPortals) {
            outgoingPortal.setDestinationDimension(linkedToPortalWorldRegKey);
            outgoingPortal.setDestination(linkedToPortalBox.getCenter());
            Vec3d axisW = Vec3d.of(PortalHelper.getAxisW(fromPortalBlockBox).getVector());
            outgoingPortal.setRotationTransformation(DQuaternion.getRotationBetween(axisW, Vec3d.of(PortalHelper.getAxisW(linkedToPortalBlockBox).getVector())));
            outgoingPortal.reloadAndSyncToClient();
        }
    }

    @Nullable
    private static ServerWorld getPortalWorld(Pair<BlockBox, Identifier> fromPortal, MinecraftServer server) {
        return server.getWorld(RegistryKey.of(RegistryKeys.WORLD, fromPortal.second));
    }

    private static void destroyPortalEntitiesInside(ServerWorld world, Box box) {
        List<Portal> portals = world.getEntitiesByType(TypeFilter.instanceOf(Portal.class), box, Entity::isAlive);
        portals.forEach(portal -> PortalAPI.removeGlobalPortal(world, portal));
    }

    public static boolean tryIgnite(ServerWorld world, BlockPos pos) {
        world.getProfiler().push("portal ignition");
        PortalListComponent portalCandidateList = PORTAL_CANDIDATE_LIST.get(world.getLevelProperties());
        PortalListComponent portalList = PORTAL_LIST.get(world.getLevelProperties());
        Set<Identifier> blockIds = portalCandidateList.getBlockIds();
        boolean ret = false;
        for (Identifier blockId : blockIds) {
            Identifier dim = world.getDimensionKey().getValue();
            List<BlockBox> portalCandidates = portalCandidateList.getContainingPortals(blockId, pos, dim);
            for (BlockBox portalCandidate : portalCandidates) {
                if (PortalHelper.isValidPortal(world, portalCandidate, blockId)) {
                    Pair<BlockBox, Identifier> next = portalCandidateList.getNext(blockId, portalCandidate, dim);
                    Pair<BlockBox, Identifier> current = Pair.of(portalCandidate, dim);
                    if (!next.equals(current)
                            && (!portalList.containsPortal(blockId, current) || !portalList.containsPortal(blockId, next))) {
                        portalList.createPortal(blockId, current);
                        portalList.createPortal(blockId, next);
                        ret = true;

                    }
                    fixPortalGroup(blockId, portalList, world.getServer());
                }
            }

        }

        world.getProfiler().pop();
        return ret;
    }
}
